/*
 *  Copyright (C) Chaperon. All rights reserved.                               
 *  ------------------------------------------------------------------------- 
 *  This software is published under the terms of the Apache Software License 
 *  version 1.1, a copy of which has been included  with this distribution in 
 *  the LICENSE file.                                                         
 */

package net.sourceforge.chaperon.grammar.production;

import java.io.Serializable;

import net.sourceforge.chaperon.grammar.Grammar;
import net.sourceforge.chaperon.grammar.symbol.NonTerminalSymbol;
import net.sourceforge.chaperon.grammar.symbol.TerminalSymbol;
import net.sourceforge.chaperon.grammar.symbol.SymbolList;
import net.sourceforge.chaperon.grammar.token.Token;
import net.sourceforge.chaperon.grammar.token.Associativity;

/**
 * This class presents a production of a grammar
 *
 * @author Stephan Michels
 * @version $Revision: 1.8 $ $Date: 2002/05/06 16:22:56 $
 */
public class Production implements Serializable, Cloneable
{

  private ProductionList    _owner      = null;

  private NonTerminalSymbol _ntsymbol   = NonTerminalSymbol.valueOf("noname");
  private SymbolList        _definition = new SymbolList();
  private ReduceType        _reducetype = ReduceType.NORMAL;

  private TerminalSymbol    _precedence = null;

  public Production()
  {
  }

  /**
   * Create a production.
   *
   * @param ntsymbol The symbol of this production
   */
  public Production(NonTerminalSymbol ntsymbol)
  {
    setSymbol(ntsymbol);
  }

  /**
   * Create a production.
   *
   * @param grammar    
   * @param ntsymbol The symbol of this production
   * @param definition The definition of this production
   */
  public Production(NonTerminalSymbol ntsymbol, SymbolList definition)
  {
    setSymbol(ntsymbol);
    setDefinition(definition);
  }

  /**
   * Create a production.
   *
   * @param ntsymbol The symbol of this production
   * @param definition The definition of this production
   * @param reducetype NONE | APPEND | REDUCE | NEGLECT
   */
  public Production(NonTerminalSymbol ntsymbol, SymbolList definition,
                    ReduceType reducetype)
  {
    setSymbol(ntsymbol);
    setDefinition(definition);
    setReduceType(reducetype);
  }

  /*public Production(Grammar grammar, Production production)
  {
    _grammar = grammar;
    _ntsymbol = production._ntsymbol; 
    _reducetype = production._reducetype;
  }*/

  /**
   *
   *
   * @param owner      
   */
  public void setOwner(ProductionList owner)
  {
    if (_owner != null)
      throw new IllegalStateException("Production has already a owner");
    _owner = owner;
  }

  /**
   *
   *
   * @return
   */
  public ProductionList getOwner()
  {
    return _owner;
  }

  /**
   * Sets the symbol for this production
   *
   * @param ntsymbol Non terminal symbol
   */
  public void setSymbol(NonTerminalSymbol ntsymbol)
  {
    if (ntsymbol == null)
      throw new NullPointerException();
    _ntsymbol = ntsymbol;
  }

  /**
   * Returns the symbol from this production
   *
   * @return Non terminal symbol
   */
  public NonTerminalSymbol getSymbol()
  {
    return _ntsymbol;
  }

  /**
   * Sets the definition for this production
   *
   * @param definition A list of symbols, which this production definies
   */
  public void setDefinition(SymbolList definition)
  {
    if (definition == null)
      throw new NullPointerException();
    _definition = definition;
  }

  /**
   * Returns the definition
   *
   * @return A list of Symbols, which the definition of this Production presented
   */
  public SymbolList getDefinition()
  {
    return _definition;
  }

  /**
   * Returns the the length of the Definition
   *
   * @return Lengt of the Definition
   */
  public int getLength()
  {
    return _definition.getSymbolCount();
  }

  /**
   * Sets the type of reduce
   *
   * @param reducetype NONE | APPEND | REDUCE | NEGLECT
   */
  public void setReduceType(ReduceType reducetype)
  {
    _reducetype = reducetype;
  }

  /**
   * Returns the type of reduce
   *
   * @return NONE | APPEND | REDUCE | NEGLECT
   */
  public ReduceType getReduceType()
  {
    return _reducetype;
  }

  /**
   * Sets the precedence from the production by a terminal symbol
   *
   * @param tsymbol Terminal symbol
   */
  public void setPrecedence(TerminalSymbol tsymbol)
  {
      _precedence = tsymbol;
  }

  public TerminalSymbol getPrecedence()
  {
    return _precedence;
  }

  public boolean hasPrecedence()
  {
    return _precedence!=null;
  }   

  /**
   * Returns the associativity from this production
   *
   * @return Associativity
   */
  public Associativity getAssociativity()
  {
    if (_owner==null)
      throw new IllegalStateException("Doesn't have a owner");
    Grammar grammar = _owner.getOwner();
    if (grammar==null)
      throw new IllegalStateException("ProductionList doesn't have a owner");

    if (_precedence != null)
      return grammar.getTokenList().getToken(_precedence).getAssociativity();
    for (int i = _definition.getSymbolCount() - 1; i >= 0; i--)
      if (_definition.getSymbol(i).isTerminal())
        return grammar.getTokenList().getToken(_definition.getSymbol(i)).getAssociativity();
    return Associativity.NONASSOC;
  }

  /**
   * Returns the priority from this production
   *
   * @return Priority
   */
  public int getPriority()
  {
    if (_owner==null)
      throw new IllegalStateException("Doesn't have a owner");
    Grammar grammar = _owner.getOwner();
    if (grammar==null)
      throw new IllegalStateException("ProductionList doesn't have a owner");

    if (_precedence != null)
      return grammar.getTokenList().getToken(_precedence).getPriority();
    for (int i = _definition.getSymbolCount() - 1; i >= 0; i--)
      if (_definition.getSymbol(i).isTerminal())
        return grammar.getTokenList().getToken(_definition.getSymbol(i)).getPriority();
    return 0;
  }

  /**
   * Return all used symbols in this production
   *
   * @return List of symbols
   */
  public SymbolList getSymbols()
  {
    SymbolList list = new SymbolList(true);

    list.addSymbol(_ntsymbol);
    list.addSymbolList(_definition);
    if (_precedence!=null)
      list.addSymbol(_precedence);
    return list;
  }

  /**
   * Converts the production to a string
   *
   * @return Content of the production
   */
  public String toString()
  {
    StringBuffer buffer = new StringBuffer();

    buffer.append(_ntsymbol + " := ");
    for (int j = 0; j < _definition.getSymbolCount(); j++)
      buffer.append(_definition.getSymbol(j) + " ");

    buffer.append("[");
    buffer.append(_reducetype);
    buffer.append("]");

    buffer.append("[prior="+getPriority()+"]");

    if (hasPrecedence())
      buffer.append("[prec="+getPrecedence()+"]");
    // else
    // buffer.append("\n");
    return buffer.toString();
  }

  public Object clone() throws CloneNotSupportedException
  {
    Production clone = new Production();

    clone._ntsymbol = _ntsymbol;
    clone._reducetype = _reducetype;
    clone._definition = (SymbolList)_definition.clone();
    clone._precedence = _precedence;

    return clone;
  }
}
