/*
 *  Copyright (C) Chaperon. All rights reserved.
 *  -------------------------------------------------------------------------
 *  This software is published under the terms of the Apache Software License
 *  version 1.1, a copy of which has been included  with this distribution in
 *  the LICENSE file.
 */

package net.sourceforge.chaperon.parser;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;

/**
 * The parser table, which the parser needs
 *
 * @author Stephan Michels
 * @version CVS $Id: ParserTable.java,v 1.7 2002/05/06 16:22:56 benedikta Exp $
 */
public class ParserTable implements Serializable
{

  // URI of the grammar
  private transient String _uri;

	// List of all terminal symbols
	private transient String[] _tsymbols;

	// Regex automates for the recognition of the terminal symbols
	private transient RegexAutomate[] _tokendefinitions;

	// List of all non terminal symbols
	private transient String[] _ntsymbols;

	// Index of the non terminal symbol for the start symbol
	private transient int _ssymbol = -1;

	// Regex automate for the recognition of a comment
	//private transient RegexAutomate _comment;

	// Regex automate for the recognition of a whitespace
	//private transient RegexAutomate _whitespace;

  /*private transient String[] _universaltokensymbols;
  private transient RegexAutomate[] _universaltokendefinitions;*/
  private transient String[] _ignorabletokensymbols;
  private transient RegexAutomate[] _ignorabletokendefinitions;

	// Indices of the non terminal symbols for the productions
	private transient byte[] _productionsymbols;

	// Lengths from the productions
	private transient byte[] _productionlengths;

	/** The production should be normal handled */
	public static final byte NORMAL = 0;

	/** The childs of the production should insert to the parent,
			if it has the same symbol */
	public static final byte APPEND = 1;

	/** The childs of the production should insert to the parent */
	public static final int RESOLVE = 2;

	/** The production should be neglected */
	public static final byte NEGLECT = 3;

	// Transitional flag for the productions
	private transient byte[] _productionreducetypes;

	// ---------- Action ------------------------

	// Actions, contains the type of action and the action argument
	private transient short[][] _actions;

	/** Error action */
	public static final byte ERROR = 0;

	/** Shift action */
	public static final byte SHIFT = 1;

	/** Reduce action */
	public static final byte REDUCE = 2;

	/** Accept action */
	public static final byte ACCEPT = 3;

	// ---------- Goto --------------------------

	// Transitions for the productions
	private transient short[][] _transitions;

	private static final long serialVersionUID = 8045275977315861215L;

	/**
	 * Constructs a empty parser table
	 *
	 * @param tsymbols Count of terminal symbols
	 * @param ntsymbols  Count of non terminal symbols
	 * @param productions Count of productions
	 * @param states Count of states
	 */
	public ParserTable(String uri, int tsymbols, int ignorabletokencount, int ntsymbols, int productions, int states)
	{
		int i, j;

    _uri = uri;

		_tsymbols = new String[tsymbols];
		for (i = 0; i < tsymbols; i++)
			_tsymbols[i] = "noname";
		_tokendefinitions = new RegexAutomate[tsymbols];
		for (i = 0; i < tsymbols; i++)
			_tokendefinitions[i] = new RegexAutomate();

    _ignorabletokensymbols = new String[ignorabletokencount];
    for (i = 0; i < ignorabletokencount; i++)
      _ignorabletokensymbols[i] = "noname";
    _ignorabletokendefinitions = new RegexAutomate[ignorabletokencount];
    for (i = 0; i < ignorabletokencount; i++)
      _ignorabletokendefinitions[i] = new RegexAutomate();

		_ntsymbols = new String[ntsymbols];
		for (i = 0; i < ntsymbols; i++)
			_ntsymbols[i] = "noname";

		_productionsymbols = new byte[productions];
		for (i = 0; i < productions; i++)
			_productionsymbols[i] = 0;
		_productionlengths = new byte[productions];
		for (i = 0; i < productions; i++)
			_productionlengths[i] = 0;
		_productionreducetypes = new byte[productions];
		for (i = 0; i < productions; i++)
			_productionreducetypes[i] = NORMAL;

		_actions = new short[states][tsymbols];
		for (i = 0; i < states; i++)
			for (j = 0; j < tsymbols; j++)
				_actions[i][j] = ERROR;
		_transitions = new short[states][ntsymbols];
		for (i = 0; i < states; i++)
			for (j = 0; j < ntsymbols; j++)
				_transitions[i][j] = 0;
	}

  /**
   * Set the uniform resource identifier of the grammar
   *
   * @param uri Uniform resource identifier of the grammar
   */
  public void setURI(String uri)
  {
    if (uri == null)
      throw new NullPointerException();
    _uri = uri; 
  } 
  
  /**
   * Return the uniform resource identifier of the grammar
   *
   * @return Uniform resource identifier of the grammar
   */
  public String getURI()
  {
    return _uri;
  }

	/**
	 * Sets the name of a terminal symbol
	 *
	 * @param index Index of the symbol
	 * @param symbol Name of the symbol
	 */
	public void setTerminalSymbol(int index, String symbol)
	{
		if ((index < 0) || (index >= _tsymbols.length))
			throw new IndexOutOfBoundsException();
		_tsymbols[index] = symbol;
	}

	/**
	 * Returns the name of a terminal symbol
	 *
	 * @param index Index of the symbol
	 *
	 * @return Name of the symbol
	 */
	public String getTerminalSymbol(int index)
	{
		if ((index < 0) || (index >= _tsymbols.length))
			throw new IndexOutOfBoundsException();
		return _tsymbols[index];
	}

	/**
	 * Returns the count of terminal symbols
	 *
	 * @return Count of terminal symbols
	 */
	public int getTerminalSymbolCount()
	{
		return _tsymbols.length;
	}

	/**
	 * Sets the regex automate for a token
	 *
	 * @param index Index of the terminal symbol
	 * @param definition Regex automate
	 */
	public void setTokenDefinition(int index, RegexAutomate definition)
	{
		if ((index < 0) || (index >= _tokendefinitions.length))
			throw new IndexOutOfBoundsException();
		_tokendefinitions[index] = definition;
	}

	/**
	 * Returns the regex automate of a token
	 *
	 * @param index Index of the terminal symbol
	 *
	 * @return Regex automate
	 */
  public RegexAutomate getTokenDefinition(int index)
	{
		if ((index < 0) || (index >= _tokendefinitions.length))
			throw new IndexOutOfBoundsException();
		return _tokendefinitions[index];
	}

	/**
	 * Sets the name of a non terminal symbol
	 *
	 * @param index Index of the symbol
	 * @param symbol Name of the symbol
	 */
	public void setNonTerminalSymbol(int index, String symbol)
	{
		if ((index < 0) || (index >= _ntsymbols.length))
			throw new IndexOutOfBoundsException();
		_ntsymbols[index] = symbol;
	}

	/**
	 * Returns the name of non treminal symbol
	 *
	 * @param index Index of the symbol
	 *
	 * @return Name of the symbol
	 */
	public String getNonTerminalSymbol(int index)
	{
		if ((index < 0) || (index >= _ntsymbols.length))
			throw new IndexOutOfBoundsException();
		return _ntsymbols[index];
	}

	/**
	 * Returns the count of non terminal symbols
	 *
	 * @return Count of non terminal symbols
	 */
	public int getNonTerminalSymbolCount()
	{
		return _ntsymbols.length;
	}

	/**
	 * Sets the symbol for a production
	 *
	 * @param index Index of the production
	 * @param symbolindex Index of the non terminal symbol
	 */
	public void setProductionSymbol(int index, int symbolindex)
	{
		if ((index < 0) || (index >= _productionsymbols.length)
						|| (symbolindex < 0) || (symbolindex >= _ntsymbols.length))
			throw new IndexOutOfBoundsException();
		_productionsymbols[index] = (byte) symbolindex;
	}

	/**
	 * Return the symbol of a production
	 *
	 * @param index Index of the production
	 *
	 * @return Index of the non terminal symbol
	 */
	public int getProductionSymbol(int index)
	{
		if ((index < 0) || (index >= _productionsymbols.length))
			throw new IndexOutOfBoundsException();
		return _productionsymbols[index];
	}

	/**
	 * Sets the length of a production
	 *
	 * @param index Index of the production
	 * @param length Length of the production
	 */
	public void setProductionLength(int index, int length)
	{
		if ((index < 0) || (index >= _productionlengths.length))
			throw new IndexOutOfBoundsException();
		if (length < 0)
			throw new IllegalArgumentException();
		_productionlengths[index] = (byte) length;
	}

	/**
	 * Return the length of a production
	 *
	 * @param index Index of a production
	 *
	 * @return Length of the production
	 */
	public int getProductionLength(int index)
	{
		if ((index < 0) || (index >= _productionlengths.length))
			throw new IndexOutOfBoundsException();
		return _productionlengths[index];
	}

	/**
	 * Sets the type of reduce of a production
	 *
	 * @param index Index of the production
	 * @param reducetype
	 */
	public void setProductionReduceType(int index, int reducetype)
	{
		if ((index < 0) || (index >= _productionreducetypes.length))
			throw new IndexOutOfBoundsException();
		if ((reducetype < NORMAL) || (reducetype > NEGLECT))
			throw new IllegalArgumentException();
		_productionreducetypes[index] = (byte) reducetype;
	}

	/**
	 * Return the type of reduce
	 *
	 * @param index Index of the production
	 *
	 * @return Transitional flag
	 */
	public int getProductionReduceType(int index)
	{
		if ((index < 0) || (index >= _productionreducetypes.length))
			throw new IndexOutOfBoundsException();
		return _productionreducetypes[index];
	}

	/**
	 * Return the count of productions
	 *
	 * @return Count of productions
	 */
	public int getProductionCount()
	{
		return _productionsymbols.length;
	}

	/**
	 * Sets the regex automate fs the comment
	 *
	 * @param automate Regex automate
	 */
	/*public void setCommentDefinition(RegexAutomate automate)
	{
		if (automate == null)
			throw new NullPointerException();
		_comment = automate;
	}*/

	/**
	 * Returns the regex automate
	 *
	 * @return Regex automate
	 */
	/*public RegexAutomate getCommentDefinition()
	{
		return _comment;
	}*/

	/**
	 * Sets the regex automate for the whitespace
	 *
	 * @param automate Regex automate
	 */
	/*public void setWhitespaceDefinition(RegexAutomate automate)
	{
		if (automate == null)
			throw new NullPointerException();
		_whitespace = automate;
	}*/

	/**
	 * Returns the regex automate for the whitespace
	 *
	 * @return Regex auotmate
	 */
	/*public RegexAutomate getWhitespaceDefinition()
	{
		return _whitespace;
	}*/

  /*public void setUniveralTokenSymbol(int index, String symbol)
  {
    if ((index < 0) || (index >= _universaltokensymbols.length))
      throw new IndexOutOfBoundsException();
    _universaltokensymbols[index] = symbol;
  }

  public String getUniveralTokenSymbol(int index)
  {
    if ((index < 0) || (index >= _universaltokensymbols.length))
      throw new IndexOutOfBoundsException();
    return _universaltokensymbols[index];
  }

  public void setUniveralTokenDefinition(int index, RegexAutomate definition)
  {
    if ((index < 0) || (index >= _universaltokendefinitions.length))
      throw new IndexOutOfBoundsException();
    _universaltokendefinitions[index] = definition;
  }

  public RegexAutomate getUniveralTokenDefinition(int index)
  {
    if ((index < 0) || (index >= _universaltokendefinitions.length))
      throw new IndexOutOfBoundsException();
    return _universaltokendefinitions[index];
  }*/

  public void setIgnorableTokenSymbol(int index, String symbol)
  {
    if ((index < 0) || (index >= _ignorabletokensymbols.length))
      throw new IndexOutOfBoundsException();
    _ignorabletokensymbols[index] = symbol;
  }

  public String getIgnorableTokenSymbol(int index)
  {
    if ((index < 0) || (index >= _ignorabletokensymbols.length))
      throw new IndexOutOfBoundsException();
    return _ignorabletokensymbols[index];
  }

  public void setIgnorableTokenDefinition(int index, RegexAutomate definition)
  {
    if ((index < 0) || (index >= _ignorabletokendefinitions.length))
      throw new IndexOutOfBoundsException();
    _ignorabletokendefinitions[index] = definition;
  }

  public RegexAutomate getIgnorableTokenDefinition(int index)
  {
    if ((index < 0) || (index >= _ignorabletokendefinitions.length))
      throw new IndexOutOfBoundsException();
    return _ignorabletokendefinitions[index];
  }

  /**
   * Returns the count of ignorable tokens
   *
   * @return Count of ignorable tokens
   */
  public int getIgnorableTokenCount()
  {
    return _ignorabletokensymbols.length;
  }

	/**
	 * Sets a error action for a given state and terminal symbol
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 * @param errorcode  Error code
	 */
	public void setErrorAction(int stateindex, int tsymbolindex, int errorcode)
	{
		if ((stateindex < 0) || (stateindex > _actions.length)
						|| (tsymbolindex < 0)
							 && (tsymbolindex >= _actions[stateindex].length))
			throw new IndexOutOfBoundsException();
		_actions[stateindex][tsymbolindex] = (short) (ERROR | (errorcode << 2));
	}

	/**
	 * Sets a error action for a given state and terminal symbol
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 * @param code Compressed instruction
	 */
	private void setAction(int stateindex, int tsymbolindex, short code)
	{
		if ((stateindex < 0) || (stateindex >= _actions.length)
						|| (tsymbolindex < 0)
						|| (tsymbolindex >= _actions[stateindex].length))
			throw new IndexOutOfBoundsException();
		_actions[stateindex][tsymbolindex] = code;
	}

	/**
	 * Sets a shift action for a given state and terminal symbol
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 * @param nextstate Destination state
	 */
	public void setShiftAction(int stateindex, int tsymbolindex, int nextstate)
	{
		if ((stateindex < 0) || (stateindex >= _actions.length)
						|| (tsymbolindex < 0)
						|| (tsymbolindex >= _actions[stateindex].length)
						|| (nextstate < 0) && (nextstate >= _actions.length))
			throw new IndexOutOfBoundsException();
		_actions[stateindex][tsymbolindex] = (short) (SHIFT | (nextstate << 2));
	}

	/**
	 * Sets a reduce action for a given state and terminal symbol
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 * @param production Index of the production, which should be reduced
	 */
	public void setReduceAction(int stateindex, int tsymbolindex,
															int production)
	{
		if ((stateindex < 0) || (stateindex >= _actions.length)
						|| (tsymbolindex < 0)
							 | (tsymbolindex >= _actions[stateindex].length))
			throw new IndexOutOfBoundsException();
		_actions[stateindex][tsymbolindex] = (short) (REDUCE | (production << 2));
	}

	/**
	 * Sets a accept action for a given state and terminal symbol
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 * @param production Index of the production, which should be reduced
	 */
	public void setAcceptAction(int stateindex, int tsymbolindex,
															int production)
	{
		if ((stateindex < 0) || (stateindex >= _actions.length)
						|| (tsymbolindex < 0)
						|| (tsymbolindex >= _actions[stateindex].length))
			throw new IndexOutOfBoundsException();
		_actions[stateindex][tsymbolindex] = (short) (ACCEPT | (production << 2));
	}

	/**
	 * Returns the action, which is given by the state and terminal symbol
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 *
	 * @return ERROR | SHIFT | REDUCE | ACCEPT
	 */
	public int getAction(int stateindex, int tsymbolindex)
	{
		if ((stateindex < 0) || (stateindex >= _actions.length)
						|| (tsymbolindex < 0)
						|| (tsymbolindex >= _actions[stateindex].length))
			throw new IndexOutOfBoundsException();
		return _actions[stateindex][tsymbolindex] & 3;
	}

	/**
	 * Test, if it is a error action
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 *
	 * @return True, if it is a error action
	 */
	public boolean isErrorAction(int stateindex, int tsymbolindex)
	{
		if ((stateindex < 0) || (stateindex >= _actions.length)
						|| (tsymbolindex < 0)
						|| (tsymbolindex >= _actions[stateindex].length))
			throw new IndexOutOfBoundsException();
		return (_actions[stateindex][tsymbolindex] & 3) == ERROR;
	}

	/**
	 * Test, if it is a shift action
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 *
	 * @return True, if it is a shift action
	 */
	public boolean isShiftAction(int stateindex, int tsymbolindex)
	{
		if ((stateindex < 0) || (stateindex >= _actions.length)
						|| (tsymbolindex < 0)
						|| (tsymbolindex >= _actions[stateindex].length))
			throw new IndexOutOfBoundsException();
		return (_actions[stateindex][tsymbolindex] & 3) == SHIFT;
	}

	/**
	 * Test, if it is a reduce action
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 *
	 * @return True, if it is a reduce action
	 */
	public boolean isReduceAction(int stateindex, int tsymbolindex)
	{
		if ((stateindex < 0) || (stateindex >= _actions.length)
						|| (tsymbolindex < 0)
						|| (tsymbolindex >= _actions[stateindex].length))
			throw new IndexOutOfBoundsException();
		return (_actions[stateindex][tsymbolindex] & 3) == REDUCE;
	}

	/**
	 * Test, if it is a accept action
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 *
	 * @return True, if it is a accept action
	 */
	public boolean isAcceptAction(int stateindex, int tsymbolindex)
	{
		if ((stateindex < 0) || (stateindex >= _actions.length)
						|| (tsymbolindex < 0)
						|| (tsymbolindex >= _actions[stateindex].length))
			throw new IndexOutOfBoundsException();
		return (_actions[stateindex][tsymbolindex] & 3) == ACCEPT;
	}

	/**
	 * Returns the argument of the action, which could be the
	 * error code, destination state, of the index of the production,
	 * which should be reduced.
	 *
	 * @param stateindex Index of the state
	 * @param tsymbolindex Index of the terminal symbol
	 *
	 * @return Argument of the Action
	 */
	public int getActionArgument(int stateindex, int tsymbolindex)
	{
		if ((stateindex < 0) || (stateindex >= _actions.length)
						|| (tsymbolindex < 0)
						|| (tsymbolindex >= _actions[stateindex].length))
			throw new IndexOutOfBoundsException();
		return _actions[stateindex][tsymbolindex] >> 2;
	}

	/**
	 * Sets transition for a given state and non terminal symbol
	 *
	 * @param stateindex Index of the state
	 * @param ntsymbolindex Index of the non terminal symbol
	 * @param nextstate Destination state
	 */
	public void setTransition(int stateindex, int ntsymbolindex, int nextstate)
	{
		if ((stateindex < 0) || (stateindex >= _transitions.length)
						|| (ntsymbolindex < 0)
						|| (ntsymbolindex >= _transitions[stateindex].length))
			throw new IndexOutOfBoundsException();
		_transitions[stateindex][ntsymbolindex] = (short) (nextstate);
	}

	/**
	 * Returns the transition for a given state and non terminal symbol
	 *
	 * @param stateindex Index of the state
	 * @param ntsymbolindex Index of the non terminal symbol
	 *
	 * @return Destination state
	 */
	public int getTransition(int stateindex, int ntsymbolindex)
	{
		if ((stateindex < 0) || (stateindex >= _transitions.length)
						|| (ntsymbolindex < 0)
						|| (ntsymbolindex >= _transitions[stateindex].length))
			throw new IndexOutOfBoundsException();
		return _transitions[stateindex][ntsymbolindex];
	}

	/**
	 * Returns the count of states
	 *
	 * @return Count of states
	 */
	public int getStateCount()
	{
		return _actions.length;
	}

	/**
	 * Serialize the parser table to a output stream
	 *
	 *
	 * @param out        
	 *
	 * @throws IOException
	 */
	private void writeObject(ObjectOutputStream out) throws IOException
	{
		out.defaultWriteObject();
		int i, j;

    out.writeUTF(_uri);

		int tsymbolcount, ignorabletokencount, ntsymbolcount, productioncount, statecount;

		out.writeInt(tsymbolcount = getTerminalSymbolCount());
    out.writeInt(ignorabletokencount = getIgnorableTokenCount());
		out.writeInt(ntsymbolcount = getNonTerminalSymbolCount());
		out.writeInt(productioncount = getProductionCount());
		out.writeInt(statecount = getStateCount());

		for (i = 0; i < tsymbolcount; i++)
			out.writeUTF(_tsymbols[i]);

		for (i = 0; i < tsymbolcount; i++)
			out.writeObject(_tokendefinitions[i]);

    for (i = 0; i < ignorabletokencount; i++)
      out.writeUTF(_ignorabletokensymbols[i]);

    for (i = 0; i < ignorabletokencount; i++)
      out.writeObject(_ignorabletokendefinitions[i]);

		for (i = 0; i < ntsymbolcount; i++)
			out.writeUTF(_ntsymbols[i]);

		for (i = 0; i < productioncount; i++)
			out.writeByte(_productionsymbols[i]);

		for (i = 0; i < productioncount; i++)
			out.writeByte(_productionlengths[i]);

		for (i = 0; i < productioncount; i++)
			out.writeByte(_productionreducetypes[i]);

		for (i = 0; i < statecount; i++)
			for (j = 0; j < tsymbolcount; j++)
				out.writeShort(_actions[i][j]);

		for (i = 0; i < statecount; i++)
			for (j = 0; j < ntsymbolcount; j++)
				out.writeShort(_transitions[i][j]);

		out.flush();
	}

	/**
	 * Reads a parser table from a input stream
	 *
	 * @param in Input stream
	 *
	 * @throws ClassNotFoundException If the class could not be found
	 * @throws IOException If a error occurs, while reading the input stream
	 */
	private void readObject(ObjectInputStream in)
					throws IOException, ClassNotFoundException
	{
		in.defaultReadObject();

		int i, j;
		char[] string;

    _uri = in.readUTF();

		int tsymbolcount = in.readInt();
    int ignorabletokencount = in.readInt();
		int ntsymbolcount = in.readInt();
		int productioncount = in.readInt();
		int statecount = in.readInt();

		_tsymbols = new String[tsymbolcount];
		_tokendefinitions = new RegexAutomate[tsymbolcount];
    _ignorabletokensymbols = new String[ignorabletokencount];
    _ignorabletokendefinitions = new RegexAutomate[ignorabletokencount];
		_ntsymbols = new String[ntsymbolcount];

		_productionsymbols = new byte[productioncount];
		_productionlengths = new byte[productioncount];
		_productionreducetypes = new byte[productioncount];

		_actions = new short[statecount][tsymbolcount];
		_transitions = new short[statecount][ntsymbolcount];

		for (i = 0; i < tsymbolcount; i++)
			_tsymbols[i] = in.readUTF();

		for (i = 0; i < tsymbolcount; i++)
			_tokendefinitions[i] = (RegexAutomate) in.readObject();

    for (i = 0; i < ignorabletokencount; i++)
      _ignorabletokensymbols[i] = in.readUTF();

    for (i = 0; i < ignorabletokencount; i++)
      _ignorabletokendefinitions[i] = (RegexAutomate) in.readObject();

		for (i = 0; i < ntsymbolcount; i++)
			_ntsymbols[i] = in.readUTF();

		for (i = 0; i < productioncount; i++)
			_productionsymbols[i] = in.readByte();

		for (i = 0; i < productioncount; i++)
			_productionlengths[i] = in.readByte();

		for (i = 0; i < productioncount; i++)
			_productionreducetypes[i] = in.readByte();

		for (i = 0; i < statecount; i++)
			for (j = 0; j < tsymbolcount; j++)
				_actions[i][j] = in.readShort();

		for (i = 0; i < statecount; i++)
			for (j = 0; j < ntsymbolcount; j++)
				_transitions[i][j] = in.readShort();
	}

	/**
	 * Converts the parser table to a string
	 *
	 * @return Content of the parser table
	 */
	public String toString()
	{
		int i, j;
		java.text.DecimalFormat format = new java.text.DecimalFormat("00");
		StringBuffer buffer = new StringBuffer();

		System.out.print("Terminal symbols:");
		for (i = 0; i < _tsymbols.length; i++)
			System.out.print(_tsymbols[i] + "(" + i + ") ");
		System.out.println();

		System.out.print("Non terminal symbols:");
		for (i = 0; i < _ntsymbols.length; i++)
			System.out.print(_ntsymbols[i] + "(" + i + ") ");
		System.out.println();

		int max = 0;

		for (i = 0; i < _tsymbols.length; i++)
			max = Math.max(max, _tsymbols[i].length());
		for (i = 0; i < _ntsymbols.length; i++)
			max = Math.max(max, _ntsymbols[i].length());

		for (i = 0; i < max; i++)
		{
			buffer.append("    ");
			for (j = 0; j < _tsymbols.length; j++)
				if ((max - i - 1) < _tsymbols[j].length())
					buffer.append(_tsymbols[j].charAt(max - i - 1) + "   ");
				else
					buffer.append("    ");

			buffer.append("  ");

			for (j = 0; j < _ntsymbols.length; j++)
				if ((max - i - 1) < _ntsymbols[j].length())
					buffer.append(_ntsymbols[j].charAt(max - i - 1) + "  ");
				else
					buffer.append("   ");

			buffer.append("\n");
		}

		for (i = 0; i < _actions.length; i++)
		{
			buffer.append(format.format(i) + "  ");
			for (j = 0; j < _actions[i].length; j++)
			{
				if ((_actions[i][j] & 3) == SHIFT)
					buffer.append("s" + format.format(_actions[i][j] >> 2) + " ");
				else if ((_actions[i][j] & 3) == REDUCE)
					buffer.append("r" + format.format(_actions[i][j] >> 2) + " ");
				else if ((_actions[i][j] & 3) == ERROR)
					buffer.append("--- ");
				else if ((_actions[i][j] & 3) == ACCEPT)
					buffer.append("ac  ");
			}

			buffer.append("| ");

			for (j = 0; j < _transitions[i].length; j++)
				if (_transitions[i][j] > 0)
					buffer.append(format.format(_transitions[i][j]) + " ");
				else
					buffer.append("-- ");
			buffer.append("\n");
		}
		return buffer.toString();
	}
}
